//--------------------------------------------------------------
// File     : sps_interpreter.c
// Datum    : 26.02.2014
// Version  : 1.0
// Autor    : UB
// EMail    : mc-4u(@)t-online.de
// Web      : www.mikrocontroller-4u.de
// CPU      : STM32F4
// IDE      : CooCox CoIDE 1.7.4
// GCC      : 4.7 2012q4
// Module   : GPIO, ADC, DMA, SPS_Parser
// Funktion : SPS-Interpreter
//            Fhrt alle Befehle aus der Opcode-Liste aus
//            bis entweder ein Fehler auftaucht oder
//            der USER die Abarbeitung beendet
//--------------------------------------------------------------


//--------------------------------------------------------------
// Includes
//--------------------------------------------------------------
#include "sps_interpreter.h"


//--------------------------------------------------------------
// Register-Adressen fuer Zeitmessung
//--------------------------------------------------------------
volatile unsigned int *DWT_CYCCNT     = (volatile unsigned int *)0xE0001004;
volatile unsigned int *DWT_CONTROL    = (volatile unsigned int *)0xE0001000;
volatile unsigned int *SCB_DEMCR      = (volatile unsigned int *)0xE000EDFC;


//--------------------------------------------------------------
// Globale Variabeln
//--------------------------------------------------------------
volatile uint32_t ADC1_DMA_Buffer[16];  // ADC-Puffer



//--------------------------------------------------------------
// interne Funktionen
//--------------------------------------------------------------
void sps_cyclus_read(void);
void sps_cyclus_execute(void);
void sps_cyclus_write(void);
void sps_cyclus_flanke(void);
void sps_cyclus_timer(void);
void sps_init_sps(void);
void sps_exec_opcode(void);
void sps_init_input(uint16_t port, uint16_t pin);
void sps_init_output(uint16_t port, uint16_t pin);
void sps_init_adc(uint16_t port, uint16_t pin, uint8_t nr);
void sps_init_adc_dma(uint8_t anzahl);
uint8_t sps_gpio_read_bit(uint16_t port, uint16_t pin);
void sps_gpio_write_bit(uint16_t port, uint16_t pin, uint8_t wert);
uint16_t sps_adc_read_word(uint32_t nr);



//--------------------------------------------------------------
// init aller Variabeln der SPS
//--------------------------------------------------------------
void Interpreter_Init(void)
{
  // init und start vom Counter fuer Zeitmessung
  *SCB_DEMCR = *SCB_DEMCR | 0x01000000;
  *DWT_CYCCNT = 0; // reset the counter
  *DWT_CONTROL = *DWT_CONTROL | 1 ; // enable the counter

  // init aller Variabeln
  sps_init_sps();
}


//--------------------------------------------------------------
// SPS einmal abarbeiten (zum Test mit Zeitmessung)
//
// Cyclus : (Read, Execute, Write, Flanke)
// ret_wert : 0=ok, >0=fehler-nr
//--------------------------------------------------------------
uint32_t Interpteter_Test(void)
{
  uint32_t ret_wert=0;
  uint32_t sps_tstart,sps_tstop,sps_time;

  // Zeit-Messung start
  sps_tstart=*DWT_CYCCNT;
  // eingange einlesen
  sps_cyclus_read();
  // sps programm abarbeiten
  sps_cyclus_execute();
  // ausgnge setzen
  sps_cyclus_write();
  // flankenmerker setzen
  sps_cyclus_flanke();
  // Zeit-Messung stop
  sps_tstop=*DWT_CYCCNT;
  sps_time=(sps_tstop-sps_tstart);
  // Zeit in us umrechnen
  SPS_GLOBAL.ist_cycluszeit_us=(sps_time/168);

  if(SPS_FILE.err_nr!=0) {
    ret_wert=SPS_FILE.err_nr;
  }
  else {
    if(SPS_GLOBAL.ist_cycluszeit_us>SPS_MAX_CYCLUS_TIME_US) {
      sps_error(SPS_ERR_HI_CYCLUS_TIME);
      ret_wert=SPS_FILE.err_nr;
    }
  }

  return(ret_wert);
}


//--------------------------------------------------------------
// SPS endlos abarbeiten (ohne Zeitmessung)
// bis esc durch user oder Error
//
// Cyclus : (Read, Execute, Write, Flanke, Timer)
// ret_wert : 0=ok, >0=fehler-nr
//--------------------------------------------------------------
uint32_t Interpteter_Start(void)
{
  uint32_t ret_wert=0;
  uint32_t ok=0;

  do {
    // warte bis Cycluszeit um ist
    while(Systick_Delay!=0);
    // Cyluszeit einstellen
    Systick_Delay=SPS_GLOBAL.soll_cycluszeit_ms;
    // eingange einlesen
    sps_cyclus_read();
    // sps programm abarbeiten
    sps_cyclus_execute();
    // ausgnge setzen
    sps_cyclus_write();
    // flankenmerker setzen
    sps_cyclus_flanke();
    // Timer aktuallisieren
    sps_cyclus_timer();
    // check auf ESC per UART
    if(UB_SPS_checkUART_ESC()!=0) ok=1;
  }
  while((ok==0) && (SPS_FILE.err_nr==0));

  if(SPS_FILE.err_nr!=0) {
    ret_wert=SPS_FILE.err_nr;
  }

  return(ret_wert);
}


//--------------------------------------------------------------
// SPS stop
// alle Ausgnge abschalten (auf Lo-Pegel)
//--------------------------------------------------------------
void Interpreter_Stop(void)
{
  uint32_t n,m;
  uint16_t adr,port,pin;

  // Ausgnge
  m=VAR_OUTPUT.anz;
  if(m>0) {
    for(n=0;n<m;n++) {
      adr=VAR_OUTPUT.address[n];
      port=(adr&0xFF00);
      pin=(adr&0x00FF);
      // Ausgang auf LO-Pegel
      sps_gpio_write_bit(port, pin, 0);
    }
  }

  // ADC und DMA ausschalten
  if(VAR_ADC.anz>0) {
    ADC_DMARequestAfterLastTransferCmd(ADC1, DISABLE);
    ADC_DMACmd(ADC1, DISABLE);
    ADC_Cmd(ADC1, DISABLE);
  }
}


//--------------------------------------------------------------
// SPS-Cyclus : read
// alle physikalischen eingnge einlesen
//--------------------------------------------------------------
void sps_cyclus_read(void)
{
  uint32_t n,m;
  uint16_t adr,port,pin;

  // eingaenge
  m=VAR_INPUT.anz;
  if(m>0) {
    for(n=0;n<m;n++) {
      adr=VAR_INPUT.address[n];
      port=(adr&0xFF00);
      pin=(adr&0x00FF);
      VAR_INPUT.bit_wert[n]=sps_gpio_read_bit(port, pin);
    }
  }

  // ADC
  m=VAR_ADC.anz;
  if(m>0) {
    for(n=0;n<m;n++) {
      VAR_ADC.word_wert[n]=sps_adc_read_word(n);
    }
  }
}


//--------------------------------------------------------------
// SPS-Cyclus : execute
// SPS programm einmal abarbeiten
//--------------------------------------------------------------
void sps_cyclus_execute(void)
{
  // init aller SPS-Variabeln
  SPS_FILE.err_nr=0;
  SPS_FILE.err_pos=0;
  SPS_FILE.stack_pos=0;
  SPS_FILE.akt_opcode=0;
  SPS_FILE.vke_bit=0;
  SPS_FILE.er_bit=0;
  SPS_FILE.or_bit=0;
  SPS_FILE.sta_bit=0;
  SPS_FILE.akku1=0;
  SPS_FILE.akku2=0;
  SPS_FILE.akt_liste=0;
  SPS_FILE.akt_nr=0;

  SPS_PROG.sps_aktpos=0;

  // Programm einmal abarbeiten
  do {
    // einen opcode ausfhren
    sps_exec_opcode();
    // naechster opcode
    SPS_PROG.sps_aktpos++;
  }while((SPS_PROG.sps_aktpos<SPS_PROG.sps_size) && (SPS_FILE.err_nr==0));
}


//--------------------------------------------------------------
// SPS-Cyclus : write
// alle physikalischen ausgnge setzen
//--------------------------------------------------------------
void sps_cyclus_write(void)
{
  uint32_t n,m;
  uint16_t adr,port,pin;
  uint8_t wert;

  // Ausgnge
  m=VAR_OUTPUT.anz;
  if(m>0) {
    for(n=0;n<m;n++) {
      adr=VAR_OUTPUT.address[n];
      port=(adr&0xFF00);
      pin=(adr&0x00FF);
      wert=VAR_OUTPUT.bit_wert[n];
      wert&=0x01;
      sps_gpio_write_bit(port, pin, wert);
    }
  }
}


//--------------------------------------------------------------
// SPS-Cyclus : flanke
// alle flankenmerker aktuallisieren
//--------------------------------------------------------------
void sps_cyclus_flanke(void)
{
  uint32_t n,m;

  // Eingaenge
  m=VAR_INPUT.anz;
  if(m>0) {
    for(n=0;n<m;n++) {
      VAR_INPUT.old_bit_wert[n]=VAR_INPUT.bit_wert[n];
    }
  }

  // Ausgaenge
  m=VAR_OUTPUT.anz;
  if(m>0) {
    for(n=0;n<m;n++) {
      VAR_OUTPUT.old_bit_wert[n]=VAR_OUTPUT.bit_wert[n];
    }
  }

  // Memory
  m=VAR_MEMORY.anz;
  if(m>0) {
    for(n=0;n<m;n++) {
      VAR_MEMORY.old_bit_wert[n]=VAR_MEMORY.bit_wert[n];
    }
  }
}


//--------------------------------------------------------------
// SPS-Cyclus : timer
// alle Timer abarbeiten
//--------------------------------------------------------------
void sps_cyclus_timer(void)
{
  uint32_t n,m;
  uint8_t typ,status;
  uint16_t wert;

  SPS_GLOBAL.ms_cnt++;
  if(SPS_GLOBAL.ms_cnt>=SPS_TIMER_TIC) {
    SPS_GLOBAL.ms_cnt=0;

    // alle Timer aktuallisieren
    m=VAR_MEMORY.anz;
    if(m>0) {
      for(n=0;n<m;n++) {
        typ=VAR_MEMORY.typ[n];
        if(typ==SPS_TYP_TIMER) {
          status=VAR_MEMORY.status[n];
          if(status==SPS_TIM_STATUS_STOP) {
            VAR_MEMORY.word_wert[n]=0;
          }
          else {
            wert=VAR_MEMORY.word_wert[n];
            if(wert>0) {
              wert--;
              VAR_MEMORY.word_wert[n]=wert;
            }
            else {
              VAR_MEMORY.status[n]=SPS_TIM_STATUS_TIMEOUT;
            }
          }
        }
      }
    }
  }
}


//--------------------------------------------------------------
// init der SPS Ein/Ausgnge und Variabeln
//--------------------------------------------------------------
void sps_init_sps(void)
{
  uint32_t n,m;
  uint16_t adr,port,pin;

  // eingaenge
  m=VAR_INPUT.anz;
  if(m>0) {
    for(n=0;n<m;n++) {
      adr=VAR_INPUT.address[n];
      port=(adr&0xFF00);
      pin=(adr&0x00FF);
      sps_init_input(port, pin);
      VAR_INPUT.bit_wert[n]=0;
      VAR_INPUT.old_bit_wert[n]=0;
    }
  }

  // Ausgnge
  m=VAR_OUTPUT.anz;
  if(m>0) {
    for(n=0;n<m;n++) {
      adr=VAR_OUTPUT.address[n];
      port=(adr&0xFF00);
      pin=(adr&0x00FF);
      sps_init_output(port, pin);
      VAR_OUTPUT.bit_wert[n]=0;
      VAR_OUTPUT.old_bit_wert[n]=0;
    }
  }

  // Memory
  m=VAR_MEMORY.anz;
  if(m>0) {
    for(n=0;n<m;n++) {
      VAR_MEMORY.bit_wert[n]=0;
      VAR_MEMORY.old_bit_wert[n]=0;
      VAR_MEMORY.word_wert[n]=0;
      VAR_MEMORY.timer_bit_wert[n]=0;
      VAR_MEMORY.status[n]=0;
    }
  }

  // ADC
  m=VAR_ADC.anz;
  if(m>0) {
    // init vom ADC1 und DMA
    sps_init_adc_dma(m);
    // init aller ADC-Eingnge
    for(n=0;n<m;n++) {
      adr=VAR_ADC.address[n];
      port=(adr&0xFF00);
      pin=(adr&0x00FF);
      sps_init_adc(port, pin,n+1);
      VAR_ADC.word_wert[n]=0;
    }
    // enable vom ADC und DMA
    ADC_DMARequestAfterLastTransferCmd(ADC1, ENABLE);
    ADC_DMACmd(ADC1, ENABLE);
    ADC_Cmd(ADC1, ENABLE);
    ADC_SoftwareStartConv(ADC1);
  }

  // Globale Variablen
  SPS_GLOBAL.ist_cycluszeit_us=0;
  SPS_GLOBAL.soll_cycluszeit_ms=SPS_SOLL_CYCLUS_TIME_MS;
  SPS_GLOBAL.ms_cnt=0;
}


//--------------------------------------------------------------
// fuehrt einen opcode aus
//--------------------------------------------------------------
void sps_exec_opcode(void)
{
  // opcode lesen
  SPS_FILE.akt_opcode=SPS_PROG.bytecode[SPS_PROG.sps_aktpos];

  // opcode ausfuehren
  switch(SPS_FILE.akt_opcode) {
    case AWL_OPCODE_U : interpreter_OC_Logic_U();break;
    case AWL_OPCODE_UK : interpreter_OC_Verzweigung_On(AWL_OPCODE_UK);break;
    case AWL_OPCODE_UN : interpreter_OC_Logic_UN();break;
    case AWL_OPCODE_UNK : interpreter_OC_Verzweigung_On(AWL_OPCODE_UNK);break;
    case AWL_OPCODE_O : interpreter_OC_Logic_O();break;
    case AWL_OPCODE_OK : interpreter_OC_Verzweigung_On(AWL_OPCODE_OK);break;
    case AWL_OPCODE_ON : interpreter_OC_Logic_ON();break;
    case AWL_OPCODE_ONK : interpreter_OC_Verzweigung_On(AWL_OPCODE_ONK);break;
    case AWL_OPCODE_X : interpreter_OC_Logic_X();break;
    case AWL_OPCODE_XK : interpreter_OC_Verzweigung_On(AWL_OPCODE_XK);break;
    case AWL_OPCODE_XN : interpreter_OC_Logic_XN();break;
    case AWL_OPCODE_XNK : interpreter_OC_Verzweigung_On(AWL_OPCODE_XNK);break;
    case AWL_OPCODE_KE : interpreter_OC_Verzweigung_Off();break;
    case AWL_OPCODE_EQ : interpreter_OC_EQUAL();break;
    case AWL_OPCODE_R : interpreter_OC_RS(AWL_OPCODE_R);break;
    case AWL_OPCODE_S : interpreter_OC_RS(AWL_OPCODE_S);break;
    case AWL_OPCODE_NOT : interpreter_OC_NOT();break;
    case AWL_OPCODE_SET : interpreter_OC_SET();break;
    case AWL_OPCODE_CLR : interpreter_OC_CLR();break;
    case AWL_OPCODE_FN : interpreter_OC_FN();break;
    case AWL_OPCODE_FP : interpreter_OC_FP();break;
    case AWL_OPCODE_VEQ : interpreter_OC_Vergleich(AWL_OPCODE_VEQ);break;
    case AWL_OPCODE_VGK : interpreter_OC_Vergleich(AWL_OPCODE_VGK);break;
    case AWL_OPCODE_VG : interpreter_OC_Vergleich(AWL_OPCODE_VG);break;
    case AWL_OPCODE_VK : interpreter_OC_Vergleich(AWL_OPCODE_VK);break;
    case AWL_OPCODE_VGG : interpreter_OC_Vergleich(AWL_OPCODE_VGG);break;
    case AWL_OPCODE_VKG : interpreter_OC_Vergleich(AWL_OPCODE_VKG);break;
    case AW_OPCODE_RA : interpreter_OC_Rechne(AW_OPCODE_RA);break;
    case AW_OPCODE_RS : interpreter_OC_Rechne(AW_OPCODE_RS);break;
    case AW_OPCODE_RM : interpreter_OC_Rechne(AW_OPCODE_RM);break;
    case AW_OPCODE_RD : interpreter_OC_Rechne(AW_OPCODE_RD);break;
    case AWL_OPCODE_LM : interpreter_OC_Load_M();break;
    case AWL_OPCODE_LC : interpreter_OC_Load_C();break;
    case AWL_OPCODE_T : interpreter_OC_Transfer();break;
    case AWL_OPCODE_ZV : interpreter_OC_Counter_ZV();break;
    case AWL_OPCODE_ZR : interpreter_OC_Counter_ZR();break;
    case AWL_OPCODE_SI : interpreter_OC_Timer_SI();break;
    case AWL_OPCODE_SV : interpreter_OC_Timer_SV();break;
    case AWL_OPCODE_SE : interpreter_OC_Timer_SE();break;
    case AWL_OPCODE_SS : interpreter_OC_Timer_SS();break;
    case AWL_OPCODE_SA : interpreter_OC_Timer_SA();break;
    default : sps_error(SPS_ERR_UNKNOWN_OPCODE);
  }
}


//--------------------------------------------------------------
// laufzeitfehler speichern
//--------------------------------------------------------------
void sps_error(uint32_t err_nr)
{
  if(SPS_FILE.err_nr==0) {
    SPS_FILE.err_nr=err_nr;
    SPS_FILE.err_pos=SPS_PROG.sps_aktpos;
  }
}


//--------------------------------------------------------------
// update von aktueller liste und nummer
//--------------------------------------------------------------
void sps_update_var(void)
{
  uint32_t akt_pos;

  akt_pos=SPS_PROG.sps_aktpos;
  SPS_FILE.akt_liste=SPS_PROG.bytecode[akt_pos+1];
  SPS_FILE.akt_nr=SPS_PROG.bytecode[akt_pos+2];
}


//--------------------------------------------------------------
// init von einem GPIO-Pin als INPUT
//--------------------------------------------------------------
void sps_init_input(uint16_t port, uint16_t pin)
{
  GPIO_InitTypeDef GPIO_InitStructure;
  GPIO_TypeDef* GPIO_PORT=GPIOA;
  uint32_t GPIO_CLK=RCC_AHB1Periph_GPIOA;
  uint16_t GPIO_PIN=99;

  if(port==SPS_PORTA) {
    GPIO_PORT=GPIOA;GPIO_CLK=RCC_AHB1Periph_GPIOA;
  }
  else if(port==SPS_PORTB) {
    GPIO_PORT=GPIOB;GPIO_CLK=RCC_AHB1Periph_GPIOB;
  }
  else if(port==SPS_PORTC) {
    GPIO_PORT=GPIOC;GPIO_CLK=RCC_AHB1Periph_GPIOC;
  }
  else if(port==SPS_PORTD) {
    GPIO_PORT=GPIOD;GPIO_CLK=RCC_AHB1Periph_GPIOD;
  }
  else if(port==SPS_PORTE) {
    GPIO_PORT=GPIOE;GPIO_CLK=RCC_AHB1Periph_GPIOE;
  }
  else if(port==SPS_PORTF) {
    GPIO_PORT=GPIOF;GPIO_CLK=RCC_AHB1Periph_GPIOF;
  }
  else if(port==SPS_PORTG) {
    GPIO_PORT=GPIOG;GPIO_CLK=RCC_AHB1Periph_GPIOG;
  }
  else {
    sps_error(SPS_ERR_UNKNOWN_PORTPIN);
    return;
  }

  if(pin==SPS_BIT0) GPIO_PIN=GPIO_Pin_0;
  if(pin==SPS_BIT1) GPIO_PIN=GPIO_Pin_1;
  if(pin==SPS_BIT2) GPIO_PIN=GPIO_Pin_2;
  if(pin==SPS_BIT3) GPIO_PIN=GPIO_Pin_3;
  if(pin==SPS_BIT4) GPIO_PIN=GPIO_Pin_4;
  if(pin==SPS_BIT5) GPIO_PIN=GPIO_Pin_5;
  if(pin==SPS_BIT6) GPIO_PIN=GPIO_Pin_6;
  if(pin==SPS_BIT7) GPIO_PIN=GPIO_Pin_7;
  if(pin==SPS_BIT8) GPIO_PIN=GPIO_Pin_8;
  if(pin==SPS_BIT9) GPIO_PIN=GPIO_Pin_9;
  if(pin==SPS_BIT10) GPIO_PIN=GPIO_Pin_10;
  if(pin==SPS_BIT11) GPIO_PIN=GPIO_Pin_11;
  if(pin==SPS_BIT12) GPIO_PIN=GPIO_Pin_12;
  if(pin==SPS_BIT13) GPIO_PIN=GPIO_Pin_13;
  if(pin==SPS_BIT14) GPIO_PIN=GPIO_Pin_14;
  if(pin==SPS_BIT15) GPIO_PIN=GPIO_Pin_15;

  if(GPIO_PIN==99) {
    sps_error(SPS_ERR_UNKNOWN_PORTPIN);
    return;
  }

  // Clock Enable
  RCC_AHB1PeriphClockCmd(GPIO_CLK, ENABLE);

  // Config als Digital-Eingang (ohne PullUp)
  GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN;
  GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
  GPIO_InitStructure.GPIO_Pin = GPIO_PIN;
  GPIO_Init(GPIO_PORT, &GPIO_InitStructure);
}


//--------------------------------------------------------------
// init von einem GPIO-Pin als OUTPUT
//--------------------------------------------------------------
void sps_init_output(uint16_t port, uint16_t pin)
{
  GPIO_InitTypeDef GPIO_InitStructure;
  GPIO_TypeDef* GPIO_PORT=GPIOA;
  uint32_t GPIO_CLK=RCC_AHB1Periph_GPIOA;
  uint16_t GPIO_PIN=99;

  if(port==SPS_PORTA) {
    GPIO_PORT=GPIOA;GPIO_CLK=RCC_AHB1Periph_GPIOA;
  }
  else if(port==SPS_PORTB) {
    GPIO_PORT=GPIOB;GPIO_CLK=RCC_AHB1Periph_GPIOB;
  }
  else if(port==SPS_PORTC) {
    GPIO_PORT=GPIOC;GPIO_CLK=RCC_AHB1Periph_GPIOC;
  }
  else if(port==SPS_PORTD) {
    GPIO_PORT=GPIOD;GPIO_CLK=RCC_AHB1Periph_GPIOD;
  }
  else if(port==SPS_PORTE) {
    GPIO_PORT=GPIOE;GPIO_CLK=RCC_AHB1Periph_GPIOE;
  }
  else if(port==SPS_PORTF) {
    GPIO_PORT=GPIOF;GPIO_CLK=RCC_AHB1Periph_GPIOF;
  }
  else if(port==SPS_PORTG) {
    GPIO_PORT=GPIOG;GPIO_CLK=RCC_AHB1Periph_GPIOG;
  }
  else {
    sps_error(SPS_ERR_UNKNOWN_PORTPIN);
    return;
  }

  if(pin==SPS_BIT0) GPIO_PIN=GPIO_Pin_0;
  if(pin==SPS_BIT1) GPIO_PIN=GPIO_Pin_1;
  if(pin==SPS_BIT2) GPIO_PIN=GPIO_Pin_2;
  if(pin==SPS_BIT3) GPIO_PIN=GPIO_Pin_3;
  if(pin==SPS_BIT4) GPIO_PIN=GPIO_Pin_4;
  if(pin==SPS_BIT5) GPIO_PIN=GPIO_Pin_5;
  if(pin==SPS_BIT6) GPIO_PIN=GPIO_Pin_6;
  if(pin==SPS_BIT7) GPIO_PIN=GPIO_Pin_7;
  if(pin==SPS_BIT8) GPIO_PIN=GPIO_Pin_8;
  if(pin==SPS_BIT9) GPIO_PIN=GPIO_Pin_9;
  if(pin==SPS_BIT10) GPIO_PIN=GPIO_Pin_10;
  if(pin==SPS_BIT11) GPIO_PIN=GPIO_Pin_11;
  if(pin==SPS_BIT12) GPIO_PIN=GPIO_Pin_12;
  if(pin==SPS_BIT13) GPIO_PIN=GPIO_Pin_13;
  if(pin==SPS_BIT14) GPIO_PIN=GPIO_Pin_14;
  if(pin==SPS_BIT15) GPIO_PIN=GPIO_Pin_15;

  if(GPIO_PIN==99) {
    sps_error(SPS_ERR_UNKNOWN_PORTPIN);
    return;
  }

  // Clock Enable
  RCC_AHB1PeriphClockCmd(GPIO_CLK, ENABLE);

  // Config als Digital-Ausgang
  GPIO_InitStructure.GPIO_Pin = GPIO_PIN;
  GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
  GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
  GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
  GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
  GPIO_Init(GPIO_PORT, &GPIO_InitStructure);

  // Lo-Pegel ausgeben
  GPIO_PORT->BSRRH = GPIO_PIN;
}


//--------------------------------------------------------------
// init von einem GPIO-Pin als ADC-INPUT
//--------------------------------------------------------------
void sps_init_adc(uint16_t port, uint16_t pin, uint8_t nr)
{
  GPIO_InitTypeDef GPIO_InitStructure;
  GPIO_TypeDef* GPIO_PORT=GPIOA;
  uint32_t GPIO_CLK=RCC_AHB1Periph_GPIOA;
  uint16_t GPIO_PIN=99;
  uint8_t adc_ch=99;

  // ADC-Kanal bestimmen
  if(port==SPS_PORTA) {
    GPIO_PORT=GPIOA;GPIO_CLK=RCC_AHB1Periph_GPIOA;
    if(pin==SPS_BIT0) adc_ch=ADC_Channel_0;
    if(pin==SPS_BIT1) adc_ch=ADC_Channel_1;
    if(pin==SPS_BIT2) adc_ch=ADC_Channel_2;
    if(pin==SPS_BIT3) adc_ch=ADC_Channel_3;
    if(pin==SPS_BIT4) adc_ch=ADC_Channel_4;
    if(pin==SPS_BIT5) adc_ch=ADC_Channel_5;
    if(pin==SPS_BIT6) adc_ch=ADC_Channel_6;
    if(pin==SPS_BIT7) adc_ch=ADC_Channel_7;
  }
  else if(port==SPS_PORTB) {
    GPIO_PORT=GPIOB;GPIO_CLK=RCC_AHB1Periph_GPIOB;
    if(pin==SPS_BIT0) adc_ch=ADC_Channel_8;
    if(pin==SPS_BIT1) adc_ch=ADC_Channel_9;
  }
  else if(port==SPS_PORTC) {
    GPIO_PORT=GPIOC;GPIO_CLK=RCC_AHB1Periph_GPIOC;
    if(pin==SPS_BIT0) adc_ch=ADC_Channel_10;
    if(pin==SPS_BIT1) adc_ch=ADC_Channel_11;
    if(pin==SPS_BIT2) adc_ch=ADC_Channel_12;
    if(pin==SPS_BIT3) adc_ch=ADC_Channel_13;
    if(pin==SPS_BIT4) adc_ch=ADC_Channel_14;
    if(pin==SPS_BIT5) adc_ch=ADC_Channel_15;
  }
  else {
    sps_error(SPS_ERR_UNKNOWN_PORTPIN);
    return;
  }

  if(pin==SPS_BIT0) GPIO_PIN=GPIO_Pin_0;
  if(pin==SPS_BIT1) GPIO_PIN=GPIO_Pin_1;
  if(pin==SPS_BIT2) GPIO_PIN=GPIO_Pin_2;
  if(pin==SPS_BIT3) GPIO_PIN=GPIO_Pin_3;
  if(pin==SPS_BIT4) GPIO_PIN=GPIO_Pin_4;
  if(pin==SPS_BIT5) GPIO_PIN=GPIO_Pin_5;
  if(pin==SPS_BIT6) GPIO_PIN=GPIO_Pin_6;
  if(pin==SPS_BIT7) GPIO_PIN=GPIO_Pin_7;

  if((GPIO_PIN==99) || (adc_ch==99)) {
    sps_error(SPS_ERR_UNKNOWN_PORTPIN);
    return;
  }

  // Clock Enable
  RCC_AHB1PeriphClockCmd(GPIO_CLK, ENABLE);

  // Config des Pins als Analog-Eingang
  GPIO_InitStructure.GPIO_Pin = GPIO_PIN;
  GPIO_InitStructure.GPIO_Mode = GPIO_Mode_AN;
  GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL ;
  GPIO_Init(GPIO_PORT, &GPIO_InitStructure);

  // zur regular Liste hinzufuegen
  ADC_RegularChannelConfig(ADC1, adc_ch, nr, ADC_SampleTime_28Cycles);
}


//--------------------------------------------------------------
// init vom ADC1 und DMA
//--------------------------------------------------------------
void sps_init_adc_dma(uint8_t anzahl)
{
  ADC_CommonInitTypeDef ADC_CommonInitStructure;
  ADC_InitTypeDef       ADC_InitStructure;
  DMA_InitTypeDef       DMA_InitStructure;

  // Clock Enable
  RCC_AHB1PeriphClockCmd(RCC_AHB1Periph_DMA2, ENABLE);

  // DMA-Disable
  DMA_Cmd(DMA2_Stream0, DISABLE);
  DMA_DeInit(DMA2_Stream0);

  // DMA2-Config
  DMA_InitStructure.DMA_Channel = DMA_Channel_0;
  DMA_InitStructure.DMA_PeripheralBaseAddr = (uint32_t)ADC1_DR_ADDRESS;
  DMA_InitStructure.DMA_Memory0BaseAddr = (uint32_t)&ADC1_DMA_Buffer;
  DMA_InitStructure.DMA_DIR = DMA_DIR_PeripheralToMemory;
  DMA_InitStructure.DMA_BufferSize = anzahl;
  DMA_InitStructure.DMA_PeripheralInc = DMA_PeripheralInc_Disable;
  DMA_InitStructure.DMA_MemoryInc = DMA_MemoryInc_Enable;
  DMA_InitStructure.DMA_PeripheralDataSize = DMA_PeripheralDataSize_Word;
  DMA_InitStructure.DMA_MemoryDataSize = DMA_MemoryDataSize_Word;
  DMA_InitStructure.DMA_Mode = DMA_Mode_Circular;
  DMA_InitStructure.DMA_Priority = DMA_Priority_High;
  DMA_InitStructure.DMA_FIFOMode = DMA_FIFOMode_Disable;
  DMA_InitStructure.DMA_FIFOThreshold = DMA_FIFOThreshold_HalfFull;
  DMA_InitStructure.DMA_MemoryBurst = DMA_MemoryBurst_Single;
  DMA_InitStructure.DMA_PeripheralBurst = DMA_PeripheralBurst_Single;
  DMA_Init(DMA2_Stream0, &DMA_InitStructure);
  DMA_Cmd(DMA2_Stream0, ENABLE);

  // ADC-Deinit
  ADC_DeInit();

  // Clock Enable
  RCC_APB2PeriphClockCmd(RCC_APB2Periph_ADC1, ENABLE);

  // ADC-Config (auf 21MHz)
  ADC_CommonInitStructure.ADC_Mode = ADC_Mode_Independent;
  ADC_CommonInitStructure.ADC_Prescaler = ADC_Prescaler_Div4;
  ADC_CommonInitStructure.ADC_DMAAccessMode = ADC_DMAAccessMode_Disabled;
  ADC_CommonInitStructure.ADC_TwoSamplingDelay = ADC_TwoSamplingDelay_5Cycles;
  ADC_CommonInit(&ADC_CommonInitStructure);

  ADC_InitStructure.ADC_Resolution = ADC_Resolution_12b;
  ADC_InitStructure.ADC_ScanConvMode = ENABLE;
  ADC_InitStructure.ADC_ContinuousConvMode = ENABLE;
  ADC_InitStructure.ADC_ExternalTrigConvEdge = ADC_ExternalTrigConvEdge_None;
  ADC_InitStructure.ADC_DataAlign = ADC_DataAlign_Right;
  ADC_InitStructure.ADC_NbrOfConversion = anzahl;
  ADC_Init(ADC1, &ADC_InitStructure);
}


//--------------------------------------------------------------
// abfragen von einem GPIO Input-Pin
// ret_wert : [0...1]
//--------------------------------------------------------------
uint8_t sps_gpio_read_bit(uint16_t port, uint16_t pin)
{
  uint8_t ret_wert=0;
  GPIO_TypeDef* GPIO_PORT=GPIOA;
  uint16_t GPIO_PIN=GPIO_Pin_0;
  BitAction wert;

  if(port==SPS_PORTA) GPIO_PORT=GPIOA;
  if(port==SPS_PORTB) GPIO_PORT=GPIOB;
  if(port==SPS_PORTC) GPIO_PORT=GPIOC;
  if(port==SPS_PORTD) GPIO_PORT=GPIOD;
  if(port==SPS_PORTE) GPIO_PORT=GPIOE;
  if(port==SPS_PORTF) GPIO_PORT=GPIOF;
  if(port==SPS_PORTG) GPIO_PORT=GPIOG;

  if(pin==SPS_BIT0) GPIO_PIN=GPIO_Pin_0;
  if(pin==SPS_BIT1) GPIO_PIN=GPIO_Pin_1;
  if(pin==SPS_BIT2) GPIO_PIN=GPIO_Pin_2;
  if(pin==SPS_BIT3) GPIO_PIN=GPIO_Pin_3;
  if(pin==SPS_BIT4) GPIO_PIN=GPIO_Pin_4;
  if(pin==SPS_BIT5) GPIO_PIN=GPIO_Pin_5;
  if(pin==SPS_BIT6) GPIO_PIN=GPIO_Pin_6;
  if(pin==SPS_BIT7) GPIO_PIN=GPIO_Pin_7;
  if(pin==SPS_BIT8) GPIO_PIN=GPIO_Pin_8;
  if(pin==SPS_BIT9) GPIO_PIN=GPIO_Pin_9;
  if(pin==SPS_BIT10) GPIO_PIN=GPIO_Pin_10;
  if(pin==SPS_BIT11) GPIO_PIN=GPIO_Pin_11;
  if(pin==SPS_BIT12) GPIO_PIN=GPIO_Pin_12;
  if(pin==SPS_BIT13) GPIO_PIN=GPIO_Pin_13;
  if(pin==SPS_BIT14) GPIO_PIN=GPIO_Pin_14;
  if(pin==SPS_BIT15) GPIO_PIN=GPIO_Pin_15;

  // pin lesen
  wert=GPIO_ReadInputDataBit(GPIO_PORT, GPIO_PIN);
  if(wert==Bit_SET) ret_wert=0x01;

  return(ret_wert);
}


//--------------------------------------------------------------
// setzen von einem GPIO Output-Pin
//--------------------------------------------------------------
void sps_gpio_write_bit(uint16_t port, uint16_t pin, uint8_t wert)
{
  GPIO_TypeDef* GPIO_PORT=GPIOA;
  uint16_t GPIO_PIN=GPIO_Pin_0;

  if(port==SPS_PORTA) GPIO_PORT=GPIOA;
  if(port==SPS_PORTB) GPIO_PORT=GPIOB;
  if(port==SPS_PORTC) GPIO_PORT=GPIOC;
  if(port==SPS_PORTD) GPIO_PORT=GPIOD;
  if(port==SPS_PORTE) GPIO_PORT=GPIOE;
  if(port==SPS_PORTF) GPIO_PORT=GPIOF;
  if(port==SPS_PORTG) GPIO_PORT=GPIOG;

  if(pin==SPS_BIT0) GPIO_PIN=GPIO_Pin_0;
  if(pin==SPS_BIT1) GPIO_PIN=GPIO_Pin_1;
  if(pin==SPS_BIT2) GPIO_PIN=GPIO_Pin_2;
  if(pin==SPS_BIT3) GPIO_PIN=GPIO_Pin_3;
  if(pin==SPS_BIT4) GPIO_PIN=GPIO_Pin_4;
  if(pin==SPS_BIT5) GPIO_PIN=GPIO_Pin_5;
  if(pin==SPS_BIT6) GPIO_PIN=GPIO_Pin_6;
  if(pin==SPS_BIT7) GPIO_PIN=GPIO_Pin_7;
  if(pin==SPS_BIT8) GPIO_PIN=GPIO_Pin_8;
  if(pin==SPS_BIT9) GPIO_PIN=GPIO_Pin_9;
  if(pin==SPS_BIT10) GPIO_PIN=GPIO_Pin_10;
  if(pin==SPS_BIT11) GPIO_PIN=GPIO_Pin_11;
  if(pin==SPS_BIT12) GPIO_PIN=GPIO_Pin_12;
  if(pin==SPS_BIT13) GPIO_PIN=GPIO_Pin_13;
  if(pin==SPS_BIT14) GPIO_PIN=GPIO_Pin_14;
  if(pin==SPS_BIT15) GPIO_PIN=GPIO_Pin_15;

  // pin setzen
  if(wert==0) {
    GPIO_PORT->BSRRH = GPIO_PIN;
  }
  else {
    GPIO_PORT->BSRRL = GPIO_PIN;
  }
}


//--------------------------------------------------------------
// lesen von einem ADC Input-Pin
// nr : 0...15
// ret_wert : [0...4095]
//--------------------------------------------------------------
uint16_t sps_adc_read_word(uint32_t nr)
{
  uint16_t ret_wert=0;

  if(nr<16) {
    ret_wert=(uint16_t)(ADC1_DMA_Buffer[nr]);
    if(ret_wert>4095) ret_wert=4095;
  }

  return(ret_wert);
}


//--------------------------------------------------------------
// aktuallisiert das Statuswort
// wert : 0 = bit loeschen
//        1 = bit setzen
//       99 = bit nicht aendern
//--------------------------------------------------------------
void sps_set_status(uint8_t er_bit, uint8_t vke_bit, uint8_t sta_bit, uint8_t or_bit)
{
  if(er_bit==0) SPS_FILE.er_bit=0;
  if(er_bit==1) SPS_FILE.er_bit=1;

  if(vke_bit==0) SPS_FILE.vke_bit=0;
  if(vke_bit==1) SPS_FILE.vke_bit=1;

  if(sta_bit==0) SPS_FILE.sta_bit=0;
  if(sta_bit==1) SPS_FILE.sta_bit=1;

  if(or_bit==0) SPS_FILE.or_bit=0;
  if(or_bit==1) SPS_FILE.or_bit=1;
}


//--------------------------------------------------------------
// bit variable einlesen
// aus (VAR_IN, VAR_OUT, VAR_MEM) liste
//--------------------------------------------------------------
uint8_t sps_read_var_bit(void)
{
  uint8_t ret_wert=0;
  uint16_t liste,nr;
  uint8_t typ;

  liste=SPS_FILE.akt_liste;
  nr=SPS_FILE.akt_nr;

  if(liste==AWL_LIST_INPUT) {
    ret_wert=VAR_INPUT.bit_wert[nr];
  }
  else if(liste==AWL_LIST_OUTPUT) {
    ret_wert=VAR_OUTPUT.bit_wert[nr];
  }
  else if(liste==AWL_LIST_MEMORY) {
    typ=VAR_MEMORY.typ[nr];
    if(typ==SPS_TYP_BOOL) {
      ret_wert=VAR_MEMORY.bit_wert[nr];
    }
    else if((typ==SPS_TYP_COUNTER) || (typ==SPS_TYP_WORD)) {
      if(VAR_MEMORY.word_wert[nr]==0) {
        ret_wert=0;
      }
      else {
        ret_wert=1;
      }
    }
    else if(typ==SPS_TYP_TIMER) {
      ret_wert=VAR_MEMORY.timer_bit_wert[nr];
    }
    else {
      sps_error(SPS_ERR_UNKNOWN_TYP);
    }
  }
  else {
    sps_error(SPS_ERR_UNKNOWN_LIST);
  }

  ret_wert&=0x01;

  return(ret_wert);
}


//--------------------------------------------------------------
// bit variable schreiben
// in (VAR_IN, VAR_OUT, VAR_MEM) liste
//--------------------------------------------------------------
void sps_write_var_bit(uint8_t wert)
{
  uint16_t liste,nr;

  liste=SPS_FILE.akt_liste;
  nr=SPS_FILE.akt_nr;

  wert&=0x01;

  if(liste==AWL_LIST_INPUT) {
    VAR_INPUT.bit_wert[nr]=wert;
  }
  else if(liste==AWL_LIST_OUTPUT) {
    VAR_OUTPUT.bit_wert[nr]=wert;
  }
  else if(liste==AWL_LIST_MEMORY) {
    VAR_MEMORY.bit_wert[nr]=wert;
  }
  else {
    sps_error(SPS_ERR_UNKNOWN_LIST);
  }
}


//--------------------------------------------------------------
// variable setzen
// in (VAR_IN, VAR_OUT, VAR_MEM) liste
//--------------------------------------------------------------
void sps_set_var(void)
{
  uint16_t liste,nr;
  uint8_t typ;

  liste=SPS_FILE.akt_liste;
  nr=SPS_FILE.akt_nr;

  if(liste==AWL_LIST_INPUT) {
    VAR_INPUT.bit_wert[nr]=1;
  }
  else if(liste==AWL_LIST_OUTPUT) {
    VAR_OUTPUT.bit_wert[nr]=1;
  }
  else if(liste==AWL_LIST_MEMORY) {
    typ=VAR_MEMORY.typ[nr];
    if(typ==SPS_TYP_BOOL) {
      VAR_MEMORY.bit_wert[nr]=1;
    }
    else if(typ==SPS_TYP_COUNTER) {
      VAR_MEMORY.word_wert[nr]=SPS_FILE.akku1;
    }
    else {
      sps_error(SPS_ERR_UNKNOWN_TYP);
    }
  }
  else {
    sps_error(SPS_ERR_UNKNOWN_LIST);
  }
}


//--------------------------------------------------------------
// variable rucksetzen
// in (VAR_IN, VAR_OUT, VAR_MEM) liste
//--------------------------------------------------------------
void sps_reset_var(void)
{
  uint16_t liste,nr;
  uint8_t typ;

  liste=SPS_FILE.akt_liste;
  nr=SPS_FILE.akt_nr;

  if(liste==AWL_LIST_INPUT) {
    VAR_INPUT.bit_wert[nr]=0;
  }
  else if(liste==AWL_LIST_OUTPUT) {
    VAR_OUTPUT.bit_wert[nr]=0;
  }
  else if(liste==AWL_LIST_MEMORY) {
    typ=VAR_MEMORY.typ[nr];
    if(typ==SPS_TYP_BOOL) {
      VAR_MEMORY.bit_wert[nr]=0;
    }
    else if(typ==SPS_TYP_COUNTER) {
      VAR_MEMORY.word_wert[nr]=0;
    }
    else if(typ==SPS_TYP_TIMER) {
      VAR_MEMORY.word_wert[nr]=0;
      VAR_MEMORY.timer_bit_wert[nr]=0;
      VAR_MEMORY.status[nr]=SPS_TIM_STATUS_STOP;
    }
    else {
      sps_error(SPS_ERR_UNKNOWN_TYP);
    }
  }
  else {
    sps_error(SPS_ERR_UNKNOWN_LIST);
  }
}


//--------------------------------------------------------------
// word variable einlesen (16bit)
// aus (VAR_MEM, VAR_ADC) liste
//--------------------------------------------------------------
uint16_t sps_read_var_word(void)
{
  uint16_t ret_wert=0;
  uint16_t liste,nr;

  liste=SPS_FILE.akt_liste;
  nr=SPS_FILE.akt_nr;

  if(liste==AWL_LIST_MEMORY) {
    ret_wert=VAR_MEMORY.word_wert[nr];
  }
  else if(liste==AWL_LIST_ADC) {
    ret_wert=VAR_ADC.word_wert[nr];
  }
  else {
    sps_error(SPS_ERR_UNKNOWN_LIST);
  }

  return(ret_wert);
}


//--------------------------------------------------------------
// word variable schreiben (16bit)
// aus (VAR_MEM) liste
//--------------------------------------------------------------
void sps_write_var_word(uint16_t wert)
{
  uint16_t liste,nr;

  liste=SPS_FILE.akt_liste;
  nr=SPS_FILE.akt_nr;

  if(liste==AWL_LIST_MEMORY) {
    VAR_MEMORY.word_wert[nr]=wert;
  }
  else {
    sps_error(SPS_ERR_UNKNOWN_LIST);
  }
}


//--------------------------------------------------------------
// testet auf Flanke oder Pegel
// in (VAR_IN, VAR_OUT, VAR_MEM) liste
//--------------------------------------------------------------
uint16_t sps_check_var_Flanke(void)
{
  uint16_t ret_wert=SPS_LO_PEGEL;
  uint16_t liste,nr;
  uint8_t old_wert=0, akt_wert=0;

  liste=SPS_FILE.akt_liste;
  nr=SPS_FILE.akt_nr;

  // aktuellen und alten wert lesen
  if(liste==AWL_LIST_INPUT) {
    akt_wert=VAR_INPUT.bit_wert[nr];
    old_wert=VAR_INPUT.old_bit_wert[nr];
  }
  else if(liste==AWL_LIST_OUTPUT) {
    akt_wert=VAR_OUTPUT.bit_wert[nr];
    old_wert=VAR_OUTPUT.old_bit_wert[nr];
  }
  else if(liste==AWL_LIST_MEMORY) {
    akt_wert=VAR_MEMORY.bit_wert[nr];
    old_wert=VAR_MEMORY.old_bit_wert[nr];
  }
  else {
    sps_error(SPS_ERR_UNKNOWN_LIST);
  }

  // flanke oder pegel bestimmen
  if(akt_wert==0) {
    if(old_wert==0) {
      ret_wert=SPS_LO_PEGEL;
    }
    else {
      ret_wert=SPS_LO_FLANKE;
    }
  }
  else {
    if(old_wert==0) {
      ret_wert=SPS_HI_FLANKE;
    }
    else {
      ret_wert=SPS_HI_PEGEL;
    }
  }

  return(ret_wert);
}


//--------------------------------------------------------------
// status wert schreiben
// aus (VAR_MEM) liste
//--------------------------------------------------------------
void sps_write_status(uint8_t wert)
{
  uint16_t liste,nr;

  liste=SPS_FILE.akt_liste;
  nr=SPS_FILE.akt_nr;

  if(liste==AWL_LIST_MEMORY) {
    if(wert==SPS_TIM_STATUS_START) {
      VAR_MEMORY.word_wert[nr]=SPS_FILE.akku1;
    }
    VAR_MEMORY.status[nr]=wert;
  }
  else {
    sps_error(SPS_ERR_UNKNOWN_LIST);
  }
}


//--------------------------------------------------------------
// status wert lesen
// aus (VAR_MEM) liste
//--------------------------------------------------------------
uint8_t sps_read_status(void)
{
  uint8_t ret_wert=0;
  uint16_t liste,nr;

  liste=SPS_FILE.akt_liste;
  nr=SPS_FILE.akt_nr;

  if(liste==AWL_LIST_MEMORY) {
    ret_wert=VAR_MEMORY.status[nr];
  }
  else {
    sps_error(SPS_ERR_UNKNOWN_LIST);
  }

  return(ret_wert);
}


//--------------------------------------------------------------
// timer bit schreiben
// aus (VAR_MEM) liste
//--------------------------------------------------------------
void sps_write_timer_bit(uint8_t wert)
{
  uint16_t liste,nr;

  liste=SPS_FILE.akt_liste;
  nr=SPS_FILE.akt_nr;

  wert&=0x01;

  if(liste==AWL_LIST_MEMORY) {
    VAR_MEMORY.timer_bit_wert[nr]=wert;
  }
  else {
    sps_error(SPS_ERR_UNKNOWN_LIST);
  }
}
